#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-02
# @Filename: logger.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import datetime
import logging
import logging.handlers
import os

from tron import g


class SDSSRotatingFileHandler(logging.handlers.BaseRotatingHandler):
    """A timed rotating file handler that rotates at a give MJD fraction.

    This class implements the timed rotating file logger handler in the SDSS
    way, i.e., rotating at a given MJD fraction (usually, MJD + 0.3 for APO
    and MJD + 0.5 for LCO). The file name includes the timestamp when it was
    created. A rollover simply closes that file and opens a new one.

    Parameters
    ----------
    filename : str
        The path to the log file.
    suffix : str
        A suffix to be added to the file when create. Can include standard
        ``strftime`` formats. Defaults to ``.%Y-%m-%dT%H:%M:%S``.
    kwargs : dict
        Arguments to be passed to `~logging.handlers.BaseRotatingHandler`.

    """

    def __init__(self, filename, suffix=None, **kwargs):

        self._filename_prefix = filename
        self.suffix = suffix or '%Y-%m-%dT%H:%M:%S'

        mode = kwargs.pop('mode', 'a')
        logging.handlers.BaseRotatingHandler.__init__(self, self._get_filename(), mode,
                                                      **kwargs)

        # MJD fraction at which to rollover.
        if g.location.lower() == 'apo':
            self.rollover_mjd = 0.3
        elif g.location.lower() == 'lco':
            self.rollover_mjd = 0.5
        else:
            self.rollover_mjd = 0.3

        self.rollAt = None
        self._set_next_rollover()

    def _get_filename(self, date=None):

        filename = self._filename_prefix

        if len(filename) > 0 and filename[-1] not in ['.', '/']:
            filename += '.'
        filename += self._get_datestring(date=date)

        return filename

    def _get_datestring(self, date=None):
        """Returns the datetime based on the suffix."""

        date = date or datetime.datetime.utcnow()

        return date.strftime(self.suffix)

    def _set_next_rollover(self):
        """Determines when to roll over next."""

        now = datetime.datetime.utcnow()

        next_rollover = datetime.datetime(now.year, now.month, now.day, 0, 0, 0)
        next_rollover += datetime.timedelta(days=self.rollover_mjd)

        # This should account for positive or negative rollover_mjd. We want
        # to be sure we don't end in a situation in which there is never another
        # rollover.
        while now > next_rollover:
            next_rollover += datetime.timedelta(days=1)

        self.rollAt = next_rollover

    def _open(self):
        """Opens the stream and creates a symlink to it."""

        stream = super(SDSSRotatingFileHandler, self)._open()

        current_name = os.path.join(os.path.dirname(self.baseFilename), 'current.log')

        if not os.path.exists(current_name):
            return stream

        current_target = os.path.realpath(os.readlink(current_name))

        if current_target != os.path.realpath(self.baseFilename):
            try:
                os.unlink(current_name)
                os.symlink(self.baseFilename, current_name)
            except BaseException:
                pass

        return stream

    def doRollover(self):
        """Do a rollover."""

        if self.stream:
            self.stream.close()
            self.stream = None

        # New file.
        dfn = self._get_filename(self.rollAt)
        if os.path.exists(dfn):
            os.remove(dfn)
        self.baseFilename = dfn

        self.stream = self._open()

        self._set_next_rollover()

    def shouldRollover(self, record):
        """Determine if rollover should occur.

        If the time is grater than ``rollAt``, rolls over.

        """

        if datetime.datetime.utcnow() > self.rollAt:
            return True

        return False


class Logfile(object):
    """Creates a new logger to log a Nub."""

    def __init__(self, name, log_dir):

        if name in logging.root.manager.loggerDict:
            # If we have already created this logger, just keep using it.
            self._logger = logging.root.manager.loggerDict[name]
        else:
            self._logger = logging.getLogger(name)
            self._logger.setLevel(1)

            log_dir = os.path.expanduser(log_dir)
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)

            self.fh = SDSSRotatingFileHandler(log_dir + '/',
                                              suffix='%Y-%m-%dT%H:%M:%S.log')
            self.fh.formatter = logging.Formatter('%(asctime)sZ %(message)s')
            self.fh.setLevel(1)

            self._logger.addHandler(self.fh)

    def log(self, txt, note='', level=1):
        self._logger.log(level, note + ' ' + txt)



__all__ = ['RawActorNub']

from tron import Misc

from .ActorNub import ActorNub
from .SocketActorNub import SocketActorNub


class RawActorNub(SocketActorNub, ActorNub):

    def getCmdForReply(self, r):
        """ Assign the current (and _only_) command ID to the reply,
        and set flag appropriately """

        try:
            activeMid = self.activeMid
        except BaseException:
            self.activeMid = activeMid = 1

        cmdDone = False
        try:
            replyText = r['RawText']
            if replyText == ' OK':
                Misc.log('rawReply', 'converting reply flag')
                r['flag'] = ':'
                cmdDone = True
                self.activeMid += 1
            else:
                Misc.log('rawReply', 'not converting reply flag :%s:' % (replyText))
        except Exception as e:
            Misc.log('rawReply', 'ignoring exceptoin: %s' % (e))

        r['cid'] = 0
        r['mid'] = activeMid
        cmd = ActorNub.getCmdForReply(self, r)

        if cmdDone:
            self.checkQueue()

        return cmd



__all__ = ['PyReplyEncoder']

import pickle

from tron import Misc
from tron.Hub.Reply.FullReply import FullReply

from .ReplyEncoder import ReplyEncoder


class PyReplyEncoder(ReplyEncoder):
    """ Encode Replys as single-line pickled python objects.
    """

    def __init__(self, **argv):
        ReplyEncoder.__init__(self, **argv)

        # How do we terminate encoded lines?
        #
        self.EOL = argv.get('EOL', '\f')
        self.encode = self.encodeFull

    def encodeFull(self, r, nub, noKeys=False):
        """ Encode a reply for a given nub.

        Encode all the information required to track the source of the command and the reply.
        """

        fullReply = FullReply()
        fullReply.initFromReply(r, noKeys)
        fullPickle = pickle.dumps(fullReply)

        if self.debug > 6:
            Misc.log('PyEncode.encode', 'encoding FullReply %s as %r' % (fullReply, fullPickle))
        elif self.debug > 3:
            Misc.log('PyEncode.encode', 'encoding FullReply %s' % (fullReply, ))

        return '%s%s' % (fullPickle, self.EOL)



#!/usr/bin/env python

__all__ = ['PollConnect']

import socket

from tron import Misc

from . import IOHandler


class PollConnect(IOHandler):
    """ Provide asynchronous socket accept() handling. """

    def __init__(self, poller, host, port, depth=5, callback=None):
        """ Set up to accept new connections on a given port.

        Args:
           poller      - the PollHandler instance to register with.
           host, port  - the host and port arguments to listen(2)
           depth       - the number of pending incoming connections to allow.
                         set to 0 to make the instance quit after one connection.
           callback    - the function to call as callback(fd, remote_addr) on new connections.

        """

        self.poller = poller
        self.depth = depth
        self.acceptMany = depth
        self.callback = callback
        if depth == 0:
            depth = 1

        self.listenFd = None
        try:
            self.listenFd = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.listenFd.bind((host, port))
            self.listenFd.listen(depth)
        except BaseException:
            if self.listenFd:
                self.listenFd.close()
            raise

        self.poller.addInput(self)

    def __del__(self):
        self.poller.removeInput(self)
        self.listenFd.close()

    def getInputFile(self):
        return self.listenFd.fileno()

    def readInput(self):

        Misc.log('IOAccept.readInput', 'accepting...')
        newfd, addr = self.listenFd.accept()

        # Listen for a single connect. Kill ourselves if we should.
        #
        if self.acceptMany == 0:
            self.poller.removeInput(self)
            self.poller = None
        self.listenFd.listen(self.depth)

        if self.callback:
            self.callback(newfd, addr)



from .ASCIIReplyDecoder import ASCIIReplyDecoder
from .PyReplyDecoder import PyReplyDecoder
from .RawReplyDecoder import RawReplyDecoder
from .ReplyDecoder import ReplyDecoder


#from BinaryReplyDecoder import *



__all__ = ['ReplyTaster']

from tron import Misc


class ReplyTaster(Misc.Object):
    """ Control which Replys we should accept. So far, we can list match against a number
        of actors and commanders.
    """

    def __init__(self, **argv):
        Misc.Object.__init__(self, **argv)

        self.actors = {}
        self.cmdrs = {}
        self.sources = {}

    def __str__(self):
        return 'ReplyTaster(actors=%s; cmdrs=%s; sources=%s)' % (list(
            self.actors.keys()), list(self.cmdrs.keys()), list(self.sources.keys()))

    def listeningTo(self):
        return list(self.actors.keys()), list(self.cmdrs.keys()), list(self.sources.keys())

    def removeFromFilter(self, actors, cmdrs, sources):
        """ Remove a list of actors and commanders to accept Replys from. """

        for i in actors:
            if i in self.actors:
                del self.actors[i]
        for c in cmdrs:
            if c in self.cmdrs:
                del self.cmdrs[c]
        for s in sources:
            if s in self.sources:
                del self.sources[s]

    def addToFilter(self, actors, cmdrs, sources):
        """ Add a list of actors and commanders to accept Replys from. """

        for i in actors:
            self.actors[i] = True
        for c in cmdrs:
            self.cmdrs[c] = True
        for s in sources:
            self.sources[s] = True

    def setFilter(self, actors, cmdrs, sources):
        """ Set the list of actors and commanders to accept Replys from. """

        self.actors = {}
        self.cmdrs = {}
        self.sources = {}

        self.addToFilter(actors, cmdrs, sources)

    def taste(self, reply):
        """ Do we accept the given Reply? """

        cmd = reply.cmd
        return cmd.cmdrName in self.cmdrs \
            or cmd.cmdrID in self.cmdrs \
            or '*' in self.sources or '*' in self.actors \
            or cmd.actorName in self.actors \
            or reply.src in self.sources



from . import cfg, strcvt
from .dates import *
from .Exceptions import *
from .hub_log import *
from .ID import *
from .LLock import *
from .location import *
from .logger import *
from .Object import *
from .qstr import *
from .tback import *
from .tcmd import *



#!/usr/bin/env python

__all__ = ['shellConnect']

import os


def shellConnect(cmd):
    """ Run a command, and return either (in_fd, out_fd, err_fd) or None. """

    try:
        fds = os.popen3(cmd, 'w')
    except BaseException:
        raise

    return fds



__all__ = ['ICCError']


class ICCError(Exception):
    """ A general exception for the ICC. Anything can throw one, passing a one line error message.
        The top-level event loop will close/cleanup/destroy any running command and return the
        error message on distxt.
    """

    def __init__(self, error, details=None):
        """ Create an ICCError.

        Args:
           error   - one line of text, intended for users. Will be returned on distxt.
           details - optional text, intended for operators/programmers.
                     Will be returned on diserrtxt.
        """

        Exception.__init__(self)

        self.error = error
        self.details = details
        if details:
            self.args = (error, details)
        else:
            self.args = (error, )



__all__ = ['ActorNub']

import re

from tron import Misc, g
from tron.Hub.Command.Command import Command

from .CoreNub import CoreNub


class ActorNub(CoreNub):
    """ Base class for ICC connections, where we send commands to and accept
    replies from the remote end.  """

    def __init__(self, poller, **argv):
        """ Create and connect an ActorNub. Takes the following optional arguments:
        grabCID    - a command to send, if we need to request a CID.
            Define our CID from the first response from the Actor
        initCmds   - list of strings
            Sent to the actor right after connection is established
        inform     - func(src, KVlist)
            If set, call when each command is dispatched and finished.
        replyCallback - func(cmd, reply)
            If set, called with eachreply line _instead of_ the cmd callback.
        logDir     - Log all I/O to the given directory
        """

        self.cid = None
        self.mid = 1

        CoreNub.__init__(self, poller, **argv)

        # Options:
        self.grabCID = argv.get('grabCID', False)
        self.initCmds = argv.get('initCmds', [])
        self.inform = argv.get('inform', None)
        self.replyCallback = argv.get('replyCallback', None)

        # Are we subject to permissions? If so, have we been given a name for
        # permissions checking, or should we just use our name?
        #
        self.needsAuth = argv.get('needsAuth', False)
        if self.needsAuth:
            self.needsAuth = self.name

        safeCmds = argv.get('safeCmds', None)
        if safeCmds is None:
            self.safeCmds = None
        else:
            # Let this fail grotesquely.
            self.safeCmds = re.compile(safeCmds)
            Misc.log('ActorNub.init', 'added safeCmds %s from %s' % (self.safeCmds, safeCmds))

        # All active commands that we are aware of, either because
        # we sent them, or because the actor replied to it.
        #
        self.liveCommands = {}

        # All active commands that we have sent to the actor.
        #
        self.ourCommands = {}

    def __str__(self):
        return 'ActorNub(%s, cid=%s, mid=%s)' % (self.ID, self.cid, self.mid)

    def connected(self):
        """ Optionally send a list of commands after the connection has been established.

        This is complicated by .grabCID. If that exists, we want to sent a command just to
        force the actor to generate our CID. If we send the init comands before that is
        established, we will not recognize the replies as replies to our init commands.
        So if .grabCID is set and our .cid is not, send the .grabCID command, and let the .grabCID
        logic in copeWithInput() call us back when we are _really_ connected. Feh.

        """

        if self.grabCID and self.cid is None:
            if isinstance(self.grabCID, str):
                initCmds = [self.grabCID]
            else:
                initCmds = []
            doRegister = False
        else:
            initCmds = self.initCmds
            doRegister = True

        Misc.log('ActorNub.connected', 'sending initCmds to %s (cid=%s)' % (self.ID, self.cid))
        for c in initCmds:
            Misc.log('ActorNub.connected', 'sending initCmd %s' % (c))
            self.sendCommand(Command('.hub', '0', g.hubMIDs.gimme(), self.name, c),
                             doRegister=doRegister)

    def sendCommand(self, c, doRegister=True):
        """ Main entry point for sending a command.

        Args:
          c   - a Command instance.

        Optional args:
          doRegister - if True (the default), keep track of replies.
        """

        # Check whether we can encode the command first:
        #
        self.__registerOurCmd(c, doRegister=doRegister)

        ec = self.encoder.encode(c)

        c.reportQueued()
        self.queueForOutput(ec)
        if self.log:
            self.log.log(ec, note='>')

    def copeWithInput(self, s):
        """ Incorporate new input: buffer it, then extract and operate on each complete reply.

        Args:
           s   - the new, but still unbuffered, input.
        """

        if self.debug > 6:
            Misc.log('Nub.copeWithInput', 'ActorNub %s read: %s' % (self.name, s))

        # Find and execute _every_ complete input.
        # The only time this function gets called is when new input comes in, so we
        # have no reliable mechanism for deferring input. If we need to, we could
        # adapt the PollHandler to query for unconsumed input. Hmm, maybe not.
        #
        while True:
            reply, leftover = self.decoder.decode(self.inputBuffer, s)
            s = None
            self.inputBuffer = leftover
            if reply is None:
                break

            if self.log:
                try:
                    txt = reply['RawText']
                except BaseException:
                    txt = 'UNKNOWN INPUT'
                self.log.log(txt, note='<')

            # Optionally try to fetch our CID by looking at the first reply.
            # The actor had better reply to our connection...
            #
            if self.cid is None and self.grabCID:
                Misc.log('Nub.copeWithInput', 'setting %s cid=%s' % (self.name, reply['cid']))
                self.cid = reply['cid']
                self.connected()

            cmd = self.getCmdForReply(reply)
            cmd.addReply(reply)

    def keyForCommand(self, cmd):
        """ Generate an immutable unique key for this command.

        This should perhaps go into the encoder?
        """

        return (str(cmd.actorCid), str(cmd.actorMid))

    def keyForReply(self, reply):
        """ Extract the key for this reply. Must match what .keyForCommand() is doing.

        This should perhaps go into the decoder?
        """

        return (str(reply['cid']), str(reply['mid']))

    def __registerCmd(self, cmd, ours):
        """ """

        key = self.keyForCommand(cmd)

        if self.debug > 0:
            Misc.log('Nub', 'registering key(ours=%s)=%s for %s' % (key, ours, cmd))
        if ours and key in self.liveCommands:
            raise RuntimeError('Duplicate command key for %s: %s' % (self, key))

        self.liveCommands[key] = cmd
        if ours:
            self.ourCommands[key] = cmd

    def __registerOurCmd(self, cmd, doRegister):
        """ We keep a registry of active commands, both ones that we sent and ones that we have detected.
        """

        cid = self.cid
        if cid is None:
            cid = 0
        cmd.connectToActor(cid, self.mid)
        self.mid += 1

        if doRegister:
            self.__registerCmd(cmd, ours=True)

    def __registerExternalCmd(self, cid, mid):
        """ Whenever we see input from a command that we did not send,
        create a Command to match. """

        key = (str(cid), str(mid))
        cmd = self.liveCommands.get(key)
        if not cmd:
            fullName = '.%s' % (self.name)
            cmd = Command(fullName,
                          '.%s' % (self.name),
                          mid,
                          self.name,
                          None,
                          actorCid=cid,
                          actorMid=mid,
                          bcastCmdInfo=False)
            self.__registerCmd(cmd, ours=False)

        return cmd

    def getCmdForReply(self, reply):
        """ Look for a command that matches the reply. Create one if none exists. """

        key = self.keyForReply(reply)

        cmd = self.liveCommands.get(key, None)
        if cmd is None and self.replyCallback is None:
            cmd = self.__registerExternalCmd(reply['cid'], reply['mid'])

        if self.flagFinishesCommand(reply['flag']):
            if cmd:
                del self.liveCommands[key]

            try:
                del self.ourCommands[key]
            except BaseException:
                pass

        return cmd

    def tasteReply(self, r):
        assert False, 'A reply was sent to an Actor(%s): %s' % (self, r)

    def statusCmd(self, cmd, doFinish=True):
        """ Send sundry status information keywords.
        """

        CoreNub.statusCmd(self, cmd, doFinish=False)
        self.listCommandsCmd(cmd, doFinish=False)

        if doFinish:
            cmd.finish()

    def listCommandsCmd(self, cmd, doFinish=True):
        """ Send command keywords.
        """

        cmd.inform('actorCmds=%s,%d,%d' %
                   (Misc.qstr(self.name), len(self.liveCommands), len(self.ourCommands)))

        for id, ourCmd in self.ourCommands.items():
            cmd.inform('actorCmd=%s,%s,%s' %
                       (Misc.qstr(self.name), Misc.qstr(id), Misc.qstr(ourCmd)))



__all__ = ['KV', 'KVDict', 'kvAsASCII']

import time
from collections import OrderedDict

from tron import Misc
from tron.Misc.cdict import cdict


""" Rethought a bit.

  All keys are stored as strings (or reprs), but may need to be extracted
  as some given type.

  Keys only exist in the context of their sources. This lets us easily extract
  all of a source's keys and also to flush all such keys when the source
  disconnects.

"""

knownEscapes = {'\r': '\\r', '\n': '\\n'}


def _doEscape(s, escape):
    """ If it exists in the string s, replace the escape string by
    an escaped version of itself. """

    if not escape:
        return s

    i = s.find(escape)
    while i >= 0:
        repl = ''.join([knownEscapes[c] for c in escape])
        s = s[:i] + repl + s[i + len(escape):]
        i = s.find(escape)

    return s


def kvAsASCII(key, val, escape=None):
    """ Return a canonical form of a keyword + value."""

    # val == None -- valueless keyword.
    if val is None:
        return str(key)

    if isinstance(val, KV):
        val = val.val

    if type(val) not in (list, tuple, type(None)):
        return '%s=%s' % (key, _doEscape(val, escape))
        # raise Exception("type(%s) for key(%s) value is not legit: %r" % (type(val), key, val))

    # "grammar" misfeature: empty lists show as "key", not as "key="
    if val is None or len(val) == 0:
        return str(key)

    values = []
    for v in val:
        if v is None:
            values.append('')
        else:
            values.append(_doEscape(v, escape))

    if values:
        return '%s=%s' % (key, ','.join(values))
    else:
        return str(key)


class KV(object):

    def __init__(self, key, val, reply):
        """ Create a single key-value variable. The key must be a string,
        and the value is either a typed value or an uninterpreted string.
        """

        self.key = key
        self.val = val
        self.reply = reply

#    def __str__(self):
#        return "%s=%s" % (self.key, self.val)

    def keyAsStr(self):
        return '%s' % (self.key)

    def valAsStr(self):
        return '%r' % (self.val)

    def valAs(self, converter):
        return converter(self.val)

    def __str__(self):
        if self.reply:
            t = self.reply.ctime
            cmd = self.reply.cmd
        else:
            t = 0.0
            cmd = None

        return 'KV(key=%s, val=%s, ctime=%0.4f, cmd=%s)' % (self.key, self.val, t, cmd)


class KVDict(Misc.Object):
    """ The main Key=Value dictionary.
    """

    def __init__(self, **argv):
        Misc.Object.__init__(self, **argv)
        self.sources = cdict(dictType=OrderedDict)

    def keyNamesForKVs(self, KVs):
        """ Return the key names for a list of raw KVs. """

        return [kv[0] for kv in KVs]

    def setKV(self, src, key, val, reply):
        """ Save the
        """

        if src is None:
            src = reply.src

        if self.debug > 5:
            Misc.log('KVDict.setKV', 'src=%r, key=%r, val=%r' % (src, key, val))

        if src not in self.sources:
            self.sources[src] = cdict(dictType=OrderedDict)

        self.sources[src][key] = KV(key, val, reply)

    def setKVsFromReply(self, reply, src=None):
        if src is None:
            src = reply.src
        self.setKVs(src, reply.KVs, reply)

    def setKVs(self, src, KVs, reply):
        if self.debug > 3:
            Misc.log('KVDict.setKVs', 'src = %r, keys = %r' % (src, KVs))

        for key, val in KVs.items():
            self.setKV(src, key, val, reply)

    def getKV(self, src, key, default=None):
        if src not in self.sources:
            return default
        val = self.sources[src].get(key, default)

        return val.val

    def getKey(self, src, key, default=None):
        if self.debug > 3:
            Misc.log('KVDict.getKey', 'get src=%s key=%s' % (src, key))

        if src not in self.sources:
            return default
        val = self.sources[src].get(key, default)

        if val == default:
            return val
        return val.val

    def addSource(self, source):
        """ Register the fact that a given source exists. """

        Misc.log('KVDict.addSource', 'adding source %s' % (source))
        if source in self.sources:
            Misc.log('KVDict.addSource', 'source %s already exists' % (source))
            return
        self.sources[source] = cdict(dictType=OrderedDict)

    def getSources(self):
        """ Return the known sources. """

        sourceList = sorted(self.sources.keys())

        return sourceList

    def clearSource(self, source):
        """ Remove all keys (well, actually the entire dictionary), for the given source.

        Args:
          source  - the name of a key source.

        Does not care if the source has no dictionary.
        """

        try:
            del self.sources[source]
        except BaseException:
            pass

    def getKeysForSource(self, source):
        """ Return all active keys for a given source.
        """
        pass

    def getValues(self, src, keys):
        """ Return an OrderedDict of values for the given list of keys.

        Args:
           src  - the key source to search.
           keys - a list of keys to fetch.

        Returns:
           - an OrderedDictionary of matched keys.
           - a list of unmatched key names.
        """

        vals = OrderedDict()

        if self.debug > 5:
            Misc.log('KVDict.getValues', 'get src=%s keys=%s' % (src, keys))

        d = self.sources.get(src, None)
        if d is None:
            return vals, keys

        if not keys:
            keys = list(d.keys())

        unmatched = []
        for k in keys:
            if k is None:
                Misc.log('getKVs', 'ignoring None key value in %r' % (keys))
                continue

            try:
                casek, val = d.fetch(k)
                vals[casek] = val
            except KeyError:
                unmatched.append(k)

        return vals, unmatched


if __name__ == '__main__':

    d = KVDict()
    #    d.setKV('hub', 'a', 1)
    d.setKVs('hub', (('b', 2), ('c', '3'), ('d', ('dfg', 123))), None)

    d.setKVs('xxx', (('b', 2), ('c', '3'), ('d', ('dfg', 4353))), None)

    print('\n'.join(map(str, d.listKVs(full=True))))
    print(d.listKVs(pattern='^hub'))
    print(d.listKVs(pattern='nomatch'))

    d.clearKeys(keys=('hub.b', 'hub.xx'))
    print(d.listKVs())

    d.clearKeys()
    print(d.listKVs())

    t0 = time.time()
    N = 100000
    for i in range(N):
        d.setKV('hub', repr(i), i * 3)
    t1 = time.time()
    KVL = d.listKVs(pattern='^1')
    t2 = time.time()

    print('%0.6fs per add' % ((t1 - t0) / N))
    print('%0.6fs per list' % ((t2 - t1) / N))



""" Various string parsers. """


def floatArgs(s, cnt=None, failWith=None):
    """ Parse a comma-delimited list of floats.

    Args:
        s        - the string to parse
        cnt      - if set, the required number of floats.
        failWith - if set, a string to flesh out error strings.

    Returns:
        a list of values.

    Raises:
        RuntimeError
    """

    try:
        stringList = s.split(',')
        floatList = list(map(float(stringList)))
    except Exception:
        if failWith:
            raise RuntimeError('%s: %s' % (failWith, s))
        else:
            raise

    if cnt is not None and len(floatList) != cnt:
        raise RuntimeError('%s. wrong number of arguments: %s' % (failWith, s))

    return floatList



import os

import tron.Misc
from tron import g, hub
from tron.Hub.Command.Encoders.ASCIICmdEncoder import ASCIICmdEncoder
from tron.Hub.Nub.SocketActorNub import SocketActorNub
from tron.Hub.Reply.Decoders.ASCIIReplyDecoder import ASCIIReplyDecoder


name = 'apogeeql'


def start(poller):
    cfg = tron.Misc.cfg.get(g.location, 'actors', doFlush=True)[name]

    stop()

    initCmds = ('ping', 'status')

    safeCmdsList = ['info', 'ping', 'version', 'status']
    safeCmds = r'^\s*({0})\s*$'.format('|'.join(safeCmdsList))

    d = ASCIIReplyDecoder(debug=3)
    e = ASCIICmdEncoder(sendCommander=True, useCID=False, debug=3)
    nub = SocketActorNub(
        poller,
        cfg['host'],
        cfg['port'],
        name=name,
        encoder=e,
        decoder=d,
        grabCID=True,  # the actor spontaneously generates a line we can eat.
        initCmds=initCmds,
        safeCmds=safeCmds,
        needsAuth=True,
        logDir=os.path.join(g.logDir, name),
        debug=3)
    hub.addActor(nub)


def stop():
    n = hub.findActor(name)
    if n:
        hub.dropActor(n)
        del n



import os
import time

from tron import g, hub
from tron.Hub.Command.Decoders.ASCIICmdDecoder import ASCIICmdDecoder
from tron.Hub.Nub.Commanders import StdinNub
from tron.Hub.Nub.Listeners import SocketListener
from tron.Hub.Reply.Encoders.ASCIIReplyEncoder import ASCIIReplyEncoder


name = 'nclient'
listenPort = 6095


def acceptStdin(in_f, out_f, addr=None):
    """ Create a command source with the given fds as input and output. """

    all = ('*', )

    nubID = g.nubIDs.gimme()
    fullname = '%s_%d' % (name, nubID)

    d = ASCIICmdDecoder(needCID=True, needMID=True, EOL='\n', hackEOL=True, name=name, debug=2)
    e = ASCIIReplyEncoder(CIDfirst=True, name=name, debug=2)
    c = StdinNub(g.poller,
                 in_f,
                 out_f,
                 name=fullname,
                 logDir=os.path.join(g.logDir, fullname),
                 encoder=e,
                 decoder=d,
                 debug=2)

    c.taster.addToFilter(all, (), all)
    hub.addCommander(c)
    time.sleep(1)


def start(poller):
    stop()

    ll = SocketListener(poller, listenPort, name, acceptStdin)
    hub.addAcceptor(ll)

    time.sleep(1)


def stop():
    ll = hub.findAcceptor(name)
    if ll:
        hub.dropAcceptor(ll)
        del ll



__all__ = ['SocketListener']

from tron import IO, Misc, hub


class SocketListener(object):
    """ Wait for connections on a given TCP port.
    """

    def __init__(self, poller, port, name, callback):

        self.name = name
        self.port = port
        self.poller = poller
        self.ID = self.name
        self.callback = callback
        self.listener = IO.PollAccept(poller, '', port, callback=self.acceptOne)

    def __del__(self):
        self.listener = None

    def shutdown(self, notifyHub=True, why=''):
        """ Release all resources and shut down.

        If called from "below" (i.e. a socket has been shutdown), just
        calls the hub's dropNub method, which will shortly call us back.

        If called from the hub, close all IO resources.
        """

        Misc.log('Hub.shutdown', 'notify=%s why=%s' % (notifyHub, why))

        if notifyHub:
            hub.dropNub(self)
        else:
            self.listener.shutdown()
            del self.listener

    def acceptOne(self, f, addr):
        f.setblocking(0)
        self.callback(f, f, addr)



__all__ = ['Object']

import sys


class Object(object):
    """ Provide a convenient place to add "global" hooks. Currently only grabs a debug=N argument.

    """

    def __init__(self, **argv):
        object.__init__(self)
        self.debug = argv.get('debug', 0)

        if self.debug > 5:
            sys.stderr.write('init %r\n' % (self))

    def __del__(self):
        if self.debug > 5:
            try:
                sys.stderr.write('del %r\n' % (self))
            except BaseException:
                pass



__all__ = ['FullReply']

from collections import OrderedDict


class FullReply(object):
    """ A single object containing all known information about a single reply. Basically flattens
    a Reply and its Command into a single structure.

    cmdrName
    cmdrMid
    cmdrCid

    actorName
    actorMid
    actorCid

    src               - who generated the flag and KVs. Usually actorName, but not always.
    flag
    KVs
    """

    def __str__(self):
        return ('FullReply(cmdrName=%s, cmdrMid=%s, cmdrCid=%s, actorName=%s, '
                'actorMid=%s, actorCid=%s, src=%s, flag=%s, KVs=%s)'
                % (self.cmdrName, self.cmdrMid, self.cmdrCid,
                   self.actorName, self.actorMid, self.actorCid,
                   self.src, self.flag, self.KVs))

    def initFromReply(self, r, noKeys=False):

        cmd = r.cmd

        self.cmdrName = cmd.cmdrName
        self.cmdrMid = cmd.cmdrMid
        self.cmdrCid = cmd.cmdrCid

        self.actorName = cmd.actorName
        self.actorMid = cmd.actorMid
        self.actorCid = cmd.actorCid

        self.src = r.src
        self.flag = r.flag

        if noKeys:
            self.KVs = OrderedDict()
        else:
            self.KVs = r.KVs



__all__ = ['CommandEncoder']

from tron import Misc


class CommandEncoder(Misc.Object):

    def __init__(self, **argv):
        Misc.Object.__init__(self, **argv)

        self.name = argv.get('name', 'unnamed')
        self.nubID = None

    def setNub(self, n):
        self.nubID = n

    def setName(self, s):
        self.name = s



__all__ = ['Command']

import re
import time
from collections import OrderedDict

from tron import Misc, Parsing, g
from tron.Hub.Reply.Reply import Reply


class Command(Misc.Object):
    """ Maintain everything required to shepherd a command through the hub.

    Basically, we need to track the commander's MID and CID, the internal XID,
    the target's ID, and the MID we use to call the target.

    The Command is created with the commander's info, and is populated with other
    things later.

    Where did the command come from?
      cmdrCID - the Commander's name, which is:
                     programName.username[.trail]
                Any real Commander will have a real programName and username.
                Synthetic Commands can be created from Replys to commands which
                we did not send. Those are assigned to .actorName
      cmdrMID - the Commander's MID
    """

    def __init__(self, cmdrID, cid, mid, tgt, cmd, **argv):
        """
        On commands from a Commander:
            - cmdrCID, cmdrMID are what the Commander sent
            - cmdrID is the programname.username (cmdr.fullname())
        On commands from an Actor (synthetic Commands):
            - cmdrCID, cmdrMID are what the Actor used.
            - cmdrID is .actorname.
        """

        Misc.Object.__init__(self, **argv)

        # An internally unique identifier for the command. We could use cmdr + the Commander's
        # CID + MID, but that would leave us trusting the Cmdr.
        self.xid = g.xids.gimme()
        self.ctime = time.time()

        # The source of the command.
        self.cmdrID = cmdrID
        try:
            cmdr = g.commanders[cmdrID]
            self.cmdrName = cmdr.name
        except BaseException:
            self.cmdrName = cmdrID
            if len(cmdrID) > 0 and cmdrID[0] != '.':
                Misc.log('Command', 'no commander %s' % (cmdrID))

        # How the caller wants the command identified.
        if cid is None:
            cid = '.%s' % (self.cmdrID)
        self.cmdrCid = cid
        self.cmdrMid = mid

        # Who will operate on the command, and the command itself.
        self.actorName = tgt
        self.cmd = cmd

        # The identification the the hub gives to the actor.
        self.actorCid = argv.get('actorCid', None)
        self.actorMid = argv.get('actorMid', 0)

        # Some Commands are essentially permanent.
        self.neverEnd = argv.get('neverEnd', False)

        # Register ourselves in a convenient place.
        # g.pendingCommands[self.xid] = self

        self.argDict = None

        # We need to put this silly test here, 'cuz the hub creates g.hubcmd at startup,
        # and g.hubcmd does not yet exist when it is being created...
        #
        self.bcastCmdInfo = argv.get('bcastCmdInfo', True)

        if g.hubcmd is not None and self.bcastCmdInfo:
            g.hubcmd.diag(
                'CmdIn=%s,%s,%s' %
                (Misc.qstr(self.cmdrCid), Misc.qstr(self.actorName), Misc.qstr(self.cmd)),
                src='cmds')

    def __str__(self):
        return 'Command(xid=%s, cmdr=%s, cmdrCid=%s, cmdrMid=%s, actor=%s, cmd=%s)' % \
               (self.xid, self.cmdrName, self.cmdrCid, self.cmdrMid, self.actorName,
                Misc.qstr(self.cmd))

    def _names(self):
        """ Return our program and username. """

        return self.cmdrName.split('.', 1)

    def username(self):
        """ Return the username that invoked us. """

        return self._names()[1]

    def program(self):
        """ Return the program that we belong to. """

        return self._names()[0]

    def reportQueued(self):
        if g.hubcmd is not None and self.bcastCmdInfo:
            g.hubcmd.diag(('CmdQueued=%d,%0.2f,%s,%s,%s,%s,%s' %
                           (self.xid, self.ctime, Misc.qstr(self.cmdrCid), self.cmdrMid,
                            Misc.qstr(self.actorName), self.actorMid, Misc.qstr(self.cmd))),
                          src='cmds')

    def connectToActor(self, cid, mid):
        """ Note the parts of the command we can only figure out when connected to the target. """

        self.actorCid = cid
        self.actorMid = mid

    r"""
       The command syntax that we accept is:
           cmd        := cmdword args
           cmdWord    := [a-zA-Z_]\S*
           args       :=
                         | arg
                         | args arg
           arg        := key | keyval
           key        := [a-zA-Z_][a-zA-Z0-9_]*
           keyval     := key\s*=\s*val
           val        := vpart
                         | val,vpart
           vpart      := string
                         | not-string
    """

    v_re = re.compile(
        r"""
        ^\s*                # Ignore leading space
        (?P<val>[^ \t,]+)   # Match the (sub-)value
        \s*                 # Ignore spaces
        (?P<rest>.*)        # Match eveything after the delimiter""", re.IGNORECASE | re.VERBOSE)

    def cmdr(self):
        """ Return our commander. """

        for c in list(g.commanders.values()):
            Misc.log('Command.cmdr()', 'checking %s in %s' % (self.cmdrName, c))
            if self.cmdrName == c.name:
                Misc.log('Command.cmdr()', 'matched %s in %s' % (self.cmdrName, c))
                return c

        Misc.log('Command.cmdr()', 'no cmdr %s in %s' % (self.cmdrName, g.commanders))
        return None

    def eatAVee(self, s):
        # Match a non-string value -- a value ended by:
        #  - whitespace,
        #  - a comma,
        #  - EOL
        #
        # Returns:
        #  - the matched value,
        #  - the rest of the input, with leading spaces removed.
        #
        # If by some misunderstanding there is no match, return the rest
        # of the line as the value. This avoids infinite loops.
        #

        # Misc.log('eatAVee', 'called with %s' % (s))

        matches = self.v_re.match(s)
        if matches:
            return matches.groupdict()
        else:
            g.hubcmd.inform('ParseError=%s' % (Misc.qstr("Consuming all trailing text '%s'" %
                                                         (s))),
                            src='hub')
        return {'val': Misc.qstr(s), 'rest': ''}

    def eatAString(self, s):
        # Match a quoting-escaped string.
        #
        # Returns:
        #  - the matched string, or the rest of the line,
        #  - the quoting level at the end of the string -- 0 if we closed
        #       the string, 1 if not.
        #  - the rest of the input, with leading spaces removed.
        #

        # Basically, scan a string while counting backslashes. If we come to a
        # quote preceded by an even number of backslashes (trivially 0), consider
        # the string matched.
        #

        # Misc.log('eatAString', 'called with %s' % (s))

        level = 0

        if len(s) == 0:
            raise SyntaxError('unexpected empty string while parsing')

        startQuote = s[0]
        if startQuote != "\"" and startQuote != "\'":
            raise SyntaxError('string does not start with a quote: %r' % (s))

        c = startQuote
        for i in range(1, len(s)):
            c = s[i]

            if c == startQuote:
                if level % 2 == 0:
                    return {'val': s[1:i], 'level': 0, 'rest': s[i + 1:].lstrip()}
            elif c == '\\':
                level += 1
            else:
                level = 0

        # OK, we fell off the end of the string without matching the closing quote.
        # Force the string to look OK so that nobody else needs to deal with a mangled string.
        #
        add = ''
        if c == '\\' and level % 2 == 1:
            add += '\\'

        g.hubcmd.inform([('ParseError', Misc.qstr('appended %s to string %s' % (add, s)))],
                        src='hub')
        s += add
        return {'val': s, 'level': 1, 'rest': ''}

    kv_re = re.compile(
        r"""
      ^\s*                          # Ignore leading space
      (?P<key>[a-z_][a-z0-9_-]*)    # Match keyword name
      (?P<delim>$|\s*=|\s+)         # Match delimiter or equals sign
      \s*                           # Ignore spaces
      (?P<rest>.*)                  # Match eveything after the delimiter""",
        re.IGNORECASE | re.VERBOSE)

    def parseKV(self, s):
        """ Try to parse a single KV.

        Return:
          {} on end-of-input
          { K {} rest-of-line } or
          { K V rest-of-line }

        """

        s = s.strip()
        if s == '':
            return None

        match = self.kv_re.match(s)
        if match is None:
            g.hubcmd.inform([('ParseError', Misc.qstr("No key-value found at '%s'" % (s)))],
                            src='hub')
            return None

        d = match.groupdict()
        Misc.log('Command.parseKV', 'kv_re=%s' % (d))

        rest = d['rest']

        if d['delim'] == '' or d['delim'][-1] != '=':
            # If the key is not delimited by and '=', we have a valueless keyword.
            #
            return {'key': d['key'], 'val': [], 'rest': rest}
        else:
            # Build up a list of comma-delimited values.
            #
            val = []

            while len(rest) != 0:
                next = rest[0]

                # Parse a (sub-)value
                #
                if next == "\"" or next == "\'":
                    dv = self.eatAString(rest)

                    if dv['level'] != 0:
                        Misc.log('parseKV', 'warning: eatAString returned with %s' % (dv))
                else:
                    dv = self.eatAVee(rest)

                if dv['val'] is not None:
                    val.append(dv['val'])
                rest = dv['rest']

                # Bail out if we:
                #   - hit EOL
                #   - have no more sub-values
                #
                if len(rest) == 0 or rest[0] != ',':
                    break

                # Keep gathering subvalues while we find commas.
                #
                rest = rest[1:].lstrip()

            return {'key': d['key'], 'val': val, 'rest': rest}

    def parse(self):
        """ Parse a raw command string into .argv """

        argv = OrderedDict()
        rest = self.cmd

        while True:
            d = self.parseKV(rest)
            if d is None:
                break

            if len(d['val']) == 0:
                val = None
            elif len(d['val']) == 1:
                val = d['val'][0]
            else:
                val = d['val']

            argv[d['key']] = val
            rest = d['rest']

        self.argv = argv

    def parseArgs(self):
        """ Parse a raw command string into an OrderedDict in .argDict. """

        if not self.argDict:
            self.argDict = Parsing.parseArgs(self.cmd)

    def match(self, opts):
        """ Searches an OrderedDict for matches.

        Args:
          argv - an OrderedDict of options.
          opts - a list of duples to match against. The duple parts are the option name
                 and a converter. If the converter is None, the option takes no argument.

        Returns:
          matches   - an OrderedDict of the matched options, with converted arguments.
          unmatched - a list of unmatched options from opts.
          leftovers - an OrderedDict of unmatched options from argv.

        Raises:
          Error     - Any parsing or conversion error.
        """

        self.parseArgs()
        return Parsing.match(self.argDict, opts)

    def coverArgs(self, requiredArgs, optionalArgs=None, ignoreFirst=None):
        """ getopt, sort of.

        Args:
           requiredArgs     - list of words which must be matched.
           optionalArgs     - list of optional words to match
           ignoreFirst      - if set, skip the first cmd word.

        Returns:
           - a dictionary of requiredArgs matches
           - a list of the requiredArgs words which were NOT matched.
           - a dictionary of optionalArgs matches
           - a list of command args not covered by requiredArgs or optionalArgs

        Notes:
           Does not return list of unmatched optionalArgs.
           Pretty much ignores argument order, for better or worse.

        command = 'jump height=14 over=cow before=6pm backwards 3 5 5'
        coverArgs(('height', 'backwards', 'withPole'), ('before', 'after')) ->
            {'height':14, 'backwards':None},
            ('withPole'),
            {'before':'6pm'},
            ('over=cow', '3', '5', '5')

        """

        if not hasattr(self, 'argv'):
            self.parse()

        Misc.log('MCCommand.coverArgs',
                 'requiredArgs=%r optionalArgs=%r ignoreFirst=%r argv=%r'
                 % (requiredArgs, optionalArgs, ignoreFirst, self.argv))

        # Start with a copy of the command args, which we consume as we copy to
        # the matched_args dict.
        #
        # Match requireArgs before optionalArgs
        #
        if requiredArgs:
            requiredArgs = list(requiredArgs)
        else:
            requiredArgs = []
        if optionalArgs:
            optionalArgs = list(optionalArgs)
        else:
            optionalArgs = []

        requiredMatches = {}
        optionalMatches = {}
        leftovers = []

        # Walk down the argument list and categorize the arguments.
        #
        for k in self.argv:
            v = self.argv[k]
            if k in requiredArgs:
                requiredMatches[k] = v
                requiredArgs.remove(k)
            elif k in optionalArgs:
                optionalMatches[k] = v
                optionalArgs.remove(k)
            else:
                leftovers.append((k, v))

        Misc.log('MCCommand.coverArgs',
                 'raw=%r requiredMatches=%r optionalMatches=%r unmatched=%r leftovers=%r'
                 % (self.argv, requiredMatches, optionalMatches, requiredArgs, leftovers))

        return requiredMatches, requiredArgs, optionalMatches, leftovers

    def addReply(self, reply, **argv):
        self.respond(reply['flag'], KVs=reply['KVs'], **argv)

    def inform(self, KVs='', **argv):
        self.makeAndSendReply('i', KVs, **argv)

    def warn(self, KVs='', **argv):
        self.makeAndSendReply('w', KVs, **argv)

    def diag(self, KVs='', **argv):
        self.makeAndSendReply('d', KVs, **argv)

    def finish(self, KVs='', **argv):
        if self.neverEnd:
            self.makeAndSendReply('i', KVs, **argv)
        else:
            self.makeAndSendReply(':', KVs, **argv)

    def fail(self, KVs='', **argv):
        if self.neverEnd:
            self.makeAndSendReply('w', KVs, **argv)
        else:
            self.makeAndSendReply('f', KVs, **argv)

    def respond(self, flag, KVs='', **argv):
        """ Respond with a specified flag. """

        self.makeAndSendReply(flag, KVs, **argv)

    def makeAndSendReply(self, flag, KVs, **argv):
        """ Bundle a flag and some KVs into a proper Reply & ship it off. """

        src = argv.get('src', self.actorName)
        bcast = argv.get('bcast', True)

        if self.debug > 0:
            Misc.log('Command.makeAndSendReply',
                     'src = %r, flag = %s, KVs = %r' % (src, flag, KVs))

        r = Reply(self, flag, KVs, src=src, bcast=bcast)
        self.reply(r, **argv)

    def reply(self, r, **argv):
        """ Finally register a Reply's KVs and offer it to any interested parties."""

        if self.debug > 1:
            Misc.log('Command.sendReply', 'reply = %r' % (r))

        if not argv.get('noRegister', False):
            g.KVs.setKVsFromReply(r)

        for c in list(g.commanders.values()):
            c.tasteReply(r)

        if r.finishesCommand():
            # del g.pendingCommands[self.xid]
            if self.bcastCmdInfo:
                g.hubcmd.diag('CmdDone=%s,%s' % (self.xid, Misc.qstr(r.flag.lower())), src='cmds')



__all__ = ['tback']

import inspect
import pprint
import sys
import traceback

from tron import Misc


def tback(system, e, info=None):
    """ Log a decently informative traceback. """

    try:
        frames = inspect.trace()
        toptrace = inspect.trace()[-1]
    except BaseException:
        one_liner = '%s: %s: %s' % (e, sys.exc_info()[0], sys.exc_info()[1])
        Misc.error(system, '======== exception botch: %s' % (one_liner))
        return

    tr_list = []
    tr_list.append('\n\n====== trace:\n')
    tr_list.append(pprint.pformat(toptrace))

    i = 0
    frames.reverse()
    for f in frames:
        # tr_list.append("\n\n====== frame %d arg+local names:\n" % (i))
        # tr_list.append(pprint.pformat(f[0].f_code.co_varnames))
        tr_list.append('\n\n====== frame %d locals:\n' % (i))
        tr_list.append(pprint.pformat(f[0].f_locals))
        i += 1

    ex_list = traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])
    Misc.error(system, '\n======== exception: %s\n' % (''.join(ex_list)))
    Misc.error(system, '\n======== exception details: %s\n' % (''.join(tr_list)))



__all__ = ['NubAuth']


import configparser
import hashlib as sha
import os

from tron import Misc, Parsing, g, hub


class NubAuth(object):
    """
    Intercepts and act on login and logout commands.
    """

    CONNECTED = 'connected'
    NOT_CONNECTED = 'not connected'
    CONNECTING = 'connecting'

    def __init__(self, **argv):
        object.__init__(self)

        self.state = self.NOT_CONNECTED
        self.nonce = None
        self.passwords = {}

    def rejectClient(self, cmd, clientType, clientVersion, clientPlatform):
        return False

    def parseVersion(self, s):
        """ Parse a comma-delimited pair of strings. Not really, though.

        I'm assuming there is no comma in the two parts of the version
        string. Oh, the shame of it. Someone should fire my sorry ass.

        """

        Misc.log('NubAut', 'parsing version %s' % (s))

        parts = s.split(',')
        dqparts = list(map(Parsing.dequote, parts))

        Misc.log('NubAut', 'parsed %s' % (dqparts))
        return dqparts

    def checkLogin(self, cmd):
        """ Try to match a name and password to an entry in the password file.

        The passwords are encrypt-only, so we need to match against the encrypted form
        of one of the password file entries. The scheme I originally chose is to allow
        left-prefix matches for the _parts_ of the program names, with '*' as a default match.
        For example, if the program given is 'PU04', we would try to find an entry in the password
        file for 'PU04', 'PU', and '*'. Basically, remove trailing digits, then try '*'.

        But that leaves holes -- enter no program, or 'XX', and you will drop through to the
        default. So always match.

        Returns True, or a string describing the problem.

        """

        if self.state != self.CONNECTING or self.nonce is None:
            return 'unexpected login ignored.'

        matched, unmatched, leftovers = cmd.match([('program', Parsing.dequote),
                                                   ('password', Parsing.dequote),
                                                   ('username', Parsing.dequote),
                                                   ('type', Parsing.dequote),
                                                   ('version', Parsing.dequote),
                                                   ('platform', Parsing.dequote)])

        if 'program' not in matched or 'password' not in matched:
            return 'not all arguments to login were found.'

        # OK. Look for the full program name:
        program = matched['program'].upper()

        if not os.path.exists(os.path.expanduser('~/.tronpass.cfg')):
            return 'password not configured in server'

        config = configparser.ConfigParser()
        config.readfp(open(os.path.expanduser('~/.tronpass.cfg')))
        ourPW = config.get('hub', program, fallback=None)

        if ourPW is None:
            return 'unknown program'

        enc = sha.sha1((self.nonce + ourPW).encode())
        enc_fallback = sha.sha1((self.nonce + program.lower()).encode())

        if ((enc.hexdigest() != matched['password']) and
                (enc_fallback.hexdigest() != matched['password'])):
            return 'incorrect password'

        # Register our IDs.
        #
        username = matched.get('username', None)
        if not username:
            username = program

        hub.dropCommander(self, doShutdown=False)
        self.setNames(program, username)
        hub.addCommander(self)

        # Try to squirrel away some useful facts about the client.
        self.clientType = matched.get('type', 'unknown')
        self.clientPlatform = matched.get('platform', 'unknown')
        self.clientVersion = matched.get('version', 'unknown')
        Misc.log('NubAuth', 'checkLogin version %s, platform %s' %
                 (self.clientVersion, self.clientPlatform))

        # Check whether we don't like the version
        #
        reject = self.rejectClient(cmd, self.clientType, self.clientVersion, self.clientPlatform)
        if reject:
            return reject

        return True

    def setUserInfo(self):
        """ Save a keyword describing our client.

        Kinda gross for it to be here. But there is no obvious good place for it,
        except maybe to name a TUINub.
        """

        Misc.log('reportAuth', 'reporting on %s' % (self))

        try:
            otherIP = self.otherIP
            otherFQDN = self.otherFQDN
        except BaseException:
            otherIP = otherFQDN = 'unknown'

        # And tell others all about us. Well, have the 'hub' tell them.
        self.userInfo = 'user=%s,%s,%s,%s,%s,%s' % \
                        (Misc.qstr(self.name), Misc.qstr(self.clientType),
                         Misc.qstr(self.clientVersion),
                         Misc.qstr(self.clientPlatform),
                         Misc.qstr(otherIP), Misc.qstr(otherFQDN))

    def makeMyNonce(self):
        """ Generate an ASCIIfied large random number. Put it in .nonce """

        import base64

        f = open('/dev/urandom', 'rb')
        bits = f.read(64)
        s = base64.encodebytes(bits)

        # base64 encoding inserts and appends NLs. Strip these.
        #
        self.nonce = s.decode().replace('\n', '')

        f.close()

    def interceptReply(self, reply):
        """ Trap and handle the login/logout commands here.

        Args:
            cmd   - the command to inspect.
        Return:
            bool  - True if we have consumed the command.
        """

        if self.state == self.CONNECTED:
            return False

        if reply.cmd.cmdrName == self.name:
            return False
        return True

    def interceptCmd(self, cmd):
        """ Trap and handle the login/logout commands here.

        Args:
            cmd   - the command to inspect.
        Return:
            bool  - True if we have consumed the command.
        """

        # Fast track the usual case.
        #
        if cmd.actorName != 'auth':
            if self.state == self.CONNECTED:
                return False
            else:
                cmd.fail('why=%s' % (Misc.qstr('please log in.')), src='auth')
                return True

        cmdWords = cmd.cmd.split()
        if len(cmdWords) < 1:
            return self.state != self.CONNECTED

        cmdWord = cmdWords[0]

        if self.state == self.CONNECTED:
            if cmdWord == 'logout':
                self.state = self.NOT_CONNECTED
                cmd.finish('bye', src='auth')
            else:
                return False
                # cmd.fail('unknownCommand=%s' % (Misc.qstr(cmdWord)),
                #         src='auth')
            return self.state != self.CONNECTED
        elif self.state == self.NOT_CONNECTED:
            if cmdWord == 'knockKnock':
                self.state = self.CONNECTING
                self.makeMyNonce()
                cmd.finish('nonce=%s' % (Misc.qstr(self.nonce)), src='auth')
            else:
                cmd.fail('why=%s' % (Misc.qstr('please log in.')), src='auth')

            return True
        else:
            if cmdWord == 'login':
                ret = self.checkLogin(cmd)
                if ret is True:
                    self.state = self.CONNECTED
                    cmd.finish(('loggedIn', 'cmdrID=%s' % Misc.qstr(self.name)), src='auth')
                    Misc.log('auth', 'logged in %s' % (self))
                    self.setUserInfo()
                    g.hubcmd.inform(self.userInfo)
                else:
                    self.state = self.NOT_CONNECTED
                    cmd.fail('why=%s' % Misc.qstr(ret), src='auth')
            else:
                self.state = self.NOT_CONNECTED
                cmd.fail('why=%s' % (Misc.qstr('please play by the rules.')), src='auth')
            return True



from tron import g, hub
from tron.Hub.Command.Decoders.ASCIICmdDecoder import ASCIICmdDecoder
from tron.Hub.Nub.Listeners import SocketListener
from tron.Hub.Nubs.StdinNub import StdinNub
from tron.Hub.Reply.Encoders.ASCIIReplyEncoder import ASCIIReplyEncoder


name = 'cmdin'
listenPort = 6098


def acceptStdin(in_f, out_f, addr=None):
    """ Create a command source with the given fds as input and output. """

    nubID = g.nubIDs.gimme()

    d = ASCIICmdDecoder(needCID=False, needMID=False, EOL='\r\n', name=name, debug=1)
    e = ASCIIReplyEncoder(name=name, simple=True, debug=1)
    c = StdinNub(g.poller,
                 in_f,
                 out_f,
                 name='%s-%d' % (name, nubID),
                 encoder=e,
                 decoder=d,
                 debug=1)
    c.taster.addToFilter(('*'), (), ('*'))
    hub.addCommander(c)


def start(poller):
    stop()

    ll = SocketListener(poller, listenPort, name, acceptStdin)
    hub.addAcceptor(ll)


def stop():
    ll = hub.findAcceptor(name)
    if ll:
        hub.dropAcceptor(ll)
        del ll



import os.path

import tron.Misc
from tron import g as hub_globals
from tron import hub
from tron.Hub.Command.Encoders.ASCIICmdEncoder import ASCIICmdEncoder
from tron.Hub.Nub.SocketActorNub import SocketActorNub
from tron.Hub.Reply.Decoders.ASCIIReplyDecoder import ASCIIReplyDecoder


name = 'tcc'


def start(poller):

    cfg = tron.Misc.cfg.get(hub_globals.location, 'actors', doFlush=True)[name]

    stop()

    isLCO = False
    try:
        isLCO = hub_globals.location == 'LCO'
    except BaseException:
        pass
    if isLCO:
        initCmds = (
            'device status',
            'show version',
        )
    else:
        initCmds = ('show version', 'show users', 'show time', 'show status', 'show inst/full',
                    'show object/full', 'show axisconfig', 'show focus', 'axis status',
                    'mir status')

    safeCmds = r'(^device )|(status$)'

    d = ASCIIReplyDecoder(EOL='\r', stripChars='\n', CIDfirst=False, debug=1)
    e = ASCIICmdEncoder(EOL='\r', useCID=False, CIDfirst=False, debug=1)
    tcc = SocketActorNub(
        poller,
        cfg['host'],
        cfg['port'],
        grabCID=True,  # Send an empty command to just get a CID
        initCmds=initCmds,
        safeCmds=safeCmds,
        needsAuth=True,
        name=name,
        encoder=e,
        decoder=d,
        logDir=os.path.join(hub_globals.logDir, name),
        debug=1)
    hub.addActor(tcc)


def stop():
    n = hub.findActor(name)
    if n:
        hub.dropActor(n)
        del n



__all__ = ['bcast']


import Vocab.InternalCmd as InternalCmd

from tron import Hub, Misc


""" A variant on instant messaging, where a Commander can inject keywords.

    The single command is:
        bcast BID name flag [keys]

    which injects the keys as coming from the given name with the MID BID

    bcast.name BID cmdrID flag keys

"""


class bcast(InternalCmd.InternalCmd):

    def __init__(self, **argv):
        InternalCmd.InternalCmd.__init__(self, 'bcast', **argv)
        self.decoder = Hub.ASCIIReplyDecoder(CIDfirst=True, debug=9)

    def sendCommand(self, cmd):
        """ Generate a Reply from the given pseudo-Command.

        cmd must have three arguments:
          - a name, which sets part of the Reply's "actor"
          - an MID to assign to the "Command"
          - a reply flag. e.g, i,w,:

        and may have other arguments, which must parse as a protocol reply. i.e.
        a semicolon-delimited list of key variables.

        Hmm, maybe the ID should be
        """

        # One interesting thing about this command is that all its arguments are
        # constructed to look like a _Reply_ instead of a command. So we use a different parser
        # from other hub words.
        #
        # What do we need to do?
        #  1) Parse the "command" line as a reply
        #
        s = cmd.cmd
        s.strip()
        s = '%s.%s\n' % ('bcast', s)
        r, leftover = self.decoder.decode(s, None)
        if not r:
            cmd.fail('bcastTxt="could not parse command line"')
            return
        if leftover:
            cmd.fail('bcastTxt=%s' %
                     (Misc.qstr('could not completely parse command line. Leftovers=%s' %
                                (leftover))))
            return

        #  2) Construct a pseudo-Command to go with it.
        c = Hub.Command(cmd.cmdrName, cmd.cmdrName, r['mid'], r['cid'], '', bcastCmdInfo=False)

        #  3) Fire it.
        c.addReply(r)

        cmd.finish('')



__all__ = ['ASCIICmdDecoder']

import re

from tron import Misc, g
from tron.Hub.Command import Command

from .CommandDecoder import CommandDecoder


class ASCIICmdDecoder(CommandDecoder):

    # REs to match commands like:
    #   cmdrName TGT command
    #
    mctc_re = re.compile(
        r"""
      \s*
      (?P<cid>[a-z_][a-z0-9_]*(\.[a-z_][a-z0-9_]*)*)
      \s+
      (?P<mid>[0-9]+)
      \s+
      (?P<tgt>[a-z_][a-z0-9_]*(\.[a-z_][a-z0-9_]*)*)
      \s+
      (?P<cmd>.*)""", re.IGNORECASE | re.VERBOSE)

    #   MID TGT command
    #
    mtc_re = re.compile(
        r"""
      \s*
      (?P<mid>[0-9]+)
      \s+
      (?P<tgt>[a-z_][a-z0-9_]*(\.[a-z_][a-z0-9_]*)*)
      \s+
      (?P<cmd>.*)""", re.IGNORECASE | re.VERBOSE)
    #   TGT command
    #
    tc_re = re.compile(
        r"""
      \s*
      (?P<tgt>[a-z_][a-z0-9_]*(\.[a-z_][a-z0-9_]*)*)
      \s+
      (?P<cmd>.*)""", re.IGNORECASE | re.VERBOSE)

    def __init__(self, **argv):

        CommandDecoder.__init__(self, **argv)

        self.EOL = argv.get('EOL', '\n')
        self.needCID = argv.get('needCID', True)
        self.needMID = argv.get('needMID', True)
        self.hackEOL = argv.get('hackEOL', False)

        if self.needCID and not self.needMID:
            Misc.log('ASCIICmdDecoder', 'if CID is needed, than MID must also be.')
        if not self.needMID:
            self.mid = 1

    def decode(self, buf, newData):
        """ Find and extract a single complete command from the given buffer.

        Returns:
           - a Command instance, or None if no complete command was found.
           - the unconsumed part of the buffer.

           If a command-sized piece is found, but cannot be parsed,
           return None, leftovers.

        """

        if newData:
            buf += newData

        eol = buf.find(self.EOL)

        if self.debug > 2:
            Misc.log('ASCIICmdDecoder.extractCmd', 'EOL at %d in buffer %r' % (eol, buf))

        # No complete command found. Return the original buffer so that the caller
        # can easily determine that no input was consumed.
        #
        if eol == -1:
            return None, buf

        # Telnet connections provide '\r\n'. Or worse, I fear.
        if self.hackEOL and len(buf) > 0:
            if buf[eol - 1] == '\r':
                self.EOL = '\r' + self.EOL
                self.hackEOL = False
                eol = buf.find(self.EOL)
                Misc.log('ASCIICmdDecoder.decode',
                         'adjusted EOL to %r (at %d) in: %r' % (self.EOL, eol, buf))
                g.hubcmd.warn(
                    'Text=%s' %
                    Misc.qstr(
                        'adjusted EOL for %s to %r (at %d) in: %r' %
                        (self.name, self.EOL, eol, buf)), src='hub')
                if eol == -1:
                    return None, buf

        cmdString = buf[:eol]
        buf = buf[eol + len(self.EOL):]

        if self.needCID:
            match = self.mctc_re.match(cmdString)
            if match is None:
                g.hubcmd.fail('ParseError=%s' %
                              (Misc.qstr('xxx Command from %s could not be parsed: %r' %
                                         (self.name, cmdString))),
                              src='hub')
                return None, buf
            d = match.groupdict()
        elif self.needMID:
            match = self.mtc_re.match(cmdString)
            if match is None:
                g.hubcmd.fail('ParseError=%s' %
                              (Misc.qstr('Command from %s could not be parsed: %r' %
                                         (self.name, cmdString))),
                              src='hub')
                return None, buf
            d = match.groupdict()
            d['cid'] = self.name
        else:
            match = self.tc_re.match(cmdString)

            mid = self.mid
            self.mid += 1

            if match is None:
                g.hubcmd.fail('ParseError' %
                              (Misc.qstr('Command from %s could not be parsed: %r' %
                                         (self.name, cmdString))),
                              src='hub')
                return None, buf
            else:
                d = match.groupdict()
                d['cid'] = self.name
                d['mid'] = str(mid)

        return Command(self.nubID, d['cid'], d['mid'], d['tgt'], d['cmd']), buf



__all__ = ['msg']

import time

import Vocab.InternalCmd as InternalCmd

from tron import Misc


""" A simple instant messaging system.

    The single command is:
        msg a bunch of text

    which responds with:
        : msgTime="2003-03-15T01:02:03" msgSender="Craig" msg="a bunch of text"

"""


class msg(InternalCmd.InternalCmd):

    def __init__(self, **argv):
        InternalCmd.InternalCmd.__init__(self, 'msg', **argv)

    def sendCommand(self, cmd):
        """ Simply arrange for the argument string to be visible by all interested commanders.
        """

        ts = time.strftime('%Y-%m-%dT%H:%M:%S', time.gmtime(cmd.ctime))
        cmd.finish('msg=%s, %s' % (Misc.qstr(ts), Misc.qstr(cmd.cmd)))



__all__ = ['BinaryCmdEncoder']

import struct

from tron import Misc

from .CommandEncoder import CommandEncoder


class BinaryCmdEncoder(CommandEncoder):

    def __init__(self, **argv):
        CommandEncoder.__init__(self, **argv)

    def encode(self, cmd):
        """ Encapsulate an old-style binary command.  """

        str = cmd.cmd
        length = len(str)
        csum = 0
        for i in range(length):
            csum ^= ord(str[i])

        # See extractCompleteCommand for better notes.
        #
        packet = struct.pack('>BBihh%dsBB' % (length), 1, 1, length, cmd.actorMid, cmd.actorCid,
                             str, csum, 4)

        if self.debug > 3:
            Misc.log('BinaryCmdEncoder', 'cmd=%s data=%r' % (cmd, packet))

        return packet



# /usr/bin/env python

__all__ = ['IOHandler']

import os
import time

from tron import Misc


class IOHandler(Misc.Object):
    """ Stub class for IO connections that can be managed by a PollHandler.

    This is the class that is called when input is available to be read and
    when output is known to be possible.

    Options:
        readSize: maximum size we read before returning to the poller.
        writeSize: max. size we write before returning to the poller.
        writeMany: if true, write as many queued item as can fit in writeSize
                   if false, only ever send a single queued item.
        in_f: the input file descriptor
        out_f: the output file descriptor.

    Bugs:
        in and out should probably not be in the same object.

    """

    def __init__(self, poller, **argv):
        Misc.Object.__init__(self, **argv)

        self.poller = poller

        Misc.log('IOHandler.init', 'IOHandler(argv=%s)' % (argv))

        # The IO size tweaks would mean something for slow network links.
        #
        self.tryToRead = argv.get('readSize', 4096)
        self.tryToWrite = argv.get('writeSize', 4096)
        self.tryToWriteMany = argv.get('writeMany', False)
        self.oneAtATime = argv.get('oneAtATime', False)

        self.in_f = self.out_f = None
        self.in_fd = self.out_fd = None
        self.outQueue = []
        self.queueLock = Misc.LLock(debug=(argv.get('debug', 0) > 7))
        self.setInputFile(argv.get('in_f', None))
        self.setOutputFile(argv.get('out_f', None))

        # Some stats
        #
        self.totalReads = 0
        self.totalBytesRead = 0
        self.largestRead = 0

        self.totalQueued = 0
        self.maxQueue = 0

        self.totalOutputs = 0
        self.totalWrites = 0
        self.totalBytesWritten = 0
        self.largestWrite = 0

    def ioshutdown(self, **argv):
        """ Unregister ourselves """

        why = argv.get('why', 'just cuz')
        Misc.log('IOhandler.ioshutdown', 'what=%s why=%s' % (self, why))

        self.setOutputFile(None)
        self.setInputFile(None)

    def shutdown(self, **argv):
        """ Unregister ourselves.

        This is meant to be overridden in a subclass.
        """

        self.ioshutdown()

    def setInputFile(self, f):
        """ Change the input file. Close and unregister any old file.
        Register the new one for input. """

        if self.debug > 2:
            Misc.log('IOHandler.setInput', '%s changing input %s to %s' % (self, self.in_f, f))

        # Detach and possibly close existing .in_f
        #
        if self.in_f is not None:
            self.poller.removeInput(self)
            if self.in_f != f:
                try:
                    self.in_f.close()
                except BaseException:
                    Misc.error('IOHandler.setInput', 'failed to close input for %s', self)

        # Establish new .in_f
        #
        self.in_f = f
        if f is None:
            self.in_fd = None
        else:
            self.in_fd = f.fileno()
            self.poller.addInput(self)

    def setOutputFile(self, f):
        """ Change the output file. Close and unregister any old file. Clear the output queue.

        This should be the only method which adjusts output registration.
        """

        if self.debug > 2:
            Misc.log('IOHandler.setOutput', '%s changing output %s to %s. queue=%s' %
                     (self, self.out_f, f, self.outQueue))

        if self.out_f is not None:
            self.poller.removeOutput(self)
            if f != self.out_f:
                try:
                    self.out_f.close()
                except BaseException:
                    Misc.error('IOHandler.setOutput', 'failed to close output for %s', self)

        # Establish new .out_f
        #
        self.out_f = f
        if f is None:
            self.out_fd = None
        else:
            self.out_fd = f.fileno()
        self.outQueue = []

    def getInputFd(self):
        """ Return the file descriptor for our input file. Called by the poller. """
        return self.in_fd

    def getOutputFd(self):
        """ Return the file descriptor for our output file. Called by the poller. """

        return self.out_fd

    def makeTimer(self, interval, callback, token):
        """ Create a timer to fire in interval seconds. """

        return self.makeTimerForTime(time.time() + interval, callback, token)

    def makeTimerForTime(self, when, callback, token):
        """ Create a timer.

        Args:
            when       - the absolute time (in ticks) to trigger.
            callback   - what function to call when triggered. called as callback(token)
            token      - an additional argument for the callback.

        Returns:
           currently a dictionary.
        """

        timer = {}
        timer['time'] = when
        timer['callback'] = callback
        timer['token'] = token

        return timer

    def addTimer(self, timer):
        self.poller.addTimer(timer)

    def queueForOutput(self, s, timer=None):
        """ Append s to the output queue. """

        assert s is not None, 'queueing nothing!'

        self.queueLock.acquire(src='queueForOutput')
        try:
            mustRegister = (self.outQueue == [])

            # Keep the output "lines" separate.
            #
            self.outQueue.append(s)

            # Add any timer.
            if timer is not None:
                self.addTimer(timer)

            # Bump the stats.
            self.totalQueued += 1
            if len(self.outQueue) > self.maxQueue:
                self.maxQueue = len(self.outQueue)

            if self.debug > 4:
                Misc.log('IOHandler.queueForOutput',
                         'appended %r to queue (len=%d) of %s' %
                         (s, len(self.outQueue), self))
            if mustRegister:
                self.poller.addOutput(self)
        finally:
            self.queueLock.release(src='queueForOutput')

    def checkQueue(self):
        """ Check whether we need to (re-) register ourselves with the poller. """

        if self.outQueue != []:
            self.poller.addOutput(self)

    def readInput(self):
        """ Read what is available to read, buffer that, and consume complete input.


        """

        error = ''
        readIn = ''
        try:
            readIn = os.read(self.in_fd, self.tryToRead).decode()
        except IOError as e:
            error = 'socket exception %s' % (e, )
            Misc.log('IOHandler.readInput', error)
            readIn = ''
        except OSError as e:
            error = 'os exception %s' % (e, )
            Misc.log('IOHandler.readInput', error)
            readIn = ''
        except Exception as e:
            error = 'unknown exception %s' % (e, )
            Misc.log('IOHandler.readInput', error)
            readIn = ''

        if self.debug > 4:
            Misc.log('IOHandler.readInput', 'read len=%d %r' % (len(readIn), readIn[:50]))

        # I/O error: by being called, we are told that we have input. But the read
        # showed no available input.
        # So close ourselves.
        #

        if readIn == '' and error == '':
            error = 'read returned nothing.'

        if error != '':
            self.shutdown(why=error)
        else:
            self.totalBytesRead += len(readIn)
            self.totalReads += 1
            if len(readIn) > self.largestRead:
                self.largestRead = len(readIn)

            self.copeWithInput(readIn)

    def mayOutput(self):
        """ Try to write as much as we should from the queue.

        We are controlled by two object variables:
            .tryToWrite: the maximum number of bytes we can
                         send before returning control to the poller.
            .tryToWriteMany: whether we can write several queued
                             commands before returning to the poller.

        We send queued items individually, regardless. It might be worth
        having a .coalesce variable to control that. I worry about the system
        limits being lower than our limits.

        """

        # Add up what we have written so far.
        totalSent = 0

        while True:

            # Try to send a single complete queued item. But truncate if we have to.
            #
            try:
                qtop = self.outQueue[0]
            except IndexError:
                self.setOutputFile(None)
                raise RuntimeError('mayOutput queue for %s is empty!' % (self))

            wlen = min(len(qtop), self.tryToWrite)
            if self.debug > 5:
                Misc.log('IOHandler.mayOutput', 'writing len=%d wlen=%d %r' %
                         (len(qtop), wlen, qtop[:min(wlen, 50)]))

            try:
                wrote = os.write(self.out_fd, qtop[:wlen].encode())
            except IOError as e:
                Misc.log('IOHandler.mayOutput', 'socket exception %r' % (e, ))
                self.shutdown(why=str(e))
                return
            except OSError as e:
                Misc.log('IOHandler.mayOutput', 'os exception %r' % (e, ))
                self.shutdown(why=str(e))
                return
            except Exception as e:
                Misc.log('IOHandler.mayOutput', 'unhandled exception %r' % (e, ))
                self.shutdown(why=str(e))
                return

            totalSent += wrote
            self.totalWrites += 1
            self.totalBytesWritten += wrote
            if wrote > self.largestWrite:
                self.largestWrite = wrote

            self.queueLock.acquire(src='mayOutput')
            try:
                # Either truncate queue[0] or remove it.
                #
                wroteFull = (wrote == len(qtop))
                if wroteFull:
                    self.outQueue.pop(0)
                else:
                    self.outQueue[0] = self.outQueue[0][wrote:]

                # Quit if we have no more to write.
                #
                if self.outQueue == []:
                    self.poller.removeOutput(self)
                    break

                # Quit if we don't want to write any more.
                if (self.oneAtATime and wroteFull):
                    break

                if self.debug > 5:
                    Misc.log('IOHandler.mayOutput', 'queue len=%d' % (len(self.outQueue)))

                # Quit if we only write one item or if we have written alot.
                #
                if self.tryToWriteMany:
                    if totalSent >= self.tryToWrite:
                        break
                else:
                    break

                self.totalOutputs += 1
            finally:
                self.queueLock.release(src='mayOutput')

    def statusCmd(self, cmd, name, doFinish=True):
        """ Send sundry status information keywords.
        """

        cmd.inform('ioConfig=%s,%d,%d,"%s"' %
                   (Misc.qstr(name),
                    self.tryToRead, self.tryToWrite, self.tryToWriteMany))
        cmd.inform('ioQueue=%s,%d,%d,%d' %
                   (Misc.qstr(name),
                    len(self.outQueue), self.totalQueued, self.maxQueue))
        cmd.inform('ioReads=%s,%d,%d,%d' %
                   (Misc.qstr(name),
                    self.totalReads, self.totalBytesRead, self.largestRead))
        cmd.inform(
            'ioWrites=%s,%d,%d,%d,%d' %
            (Misc.qstr(name),
             self.totalOutputs,
             self.totalWrites,
             self.totalBytesWritten,
             self.largestWrite))
        if doFinish:
            cmd.finish()
